#!/usr/bin/perl

# Copyright (C) 2000 Jay Beale
# Licensed under the GNU General Public License

##########################################
## Bastille Linux Text User Interface   ##
##########################################

#
# TO DO:
#
#  1) Use the newly-release Curses::Widgets 1.1 to get better text windows
#  2) Finish documenting program
#  3) Rewrite &Ask subroutine: either use Curses::Forms or just make cleaner
#  4) 

##########################################################################
# This is the Text User Interface for Bastille Linux Architecture 1,     #
# which is used in Bastille Linux 1.0-1.2.                               #
#                                                                        #
# Bastille Linux's first architecture was primarily streams-based, as it #
# was intended to be a fully interactive Perl script.  Since then, I've  #
# understood that this original interface could be drastically improved, #
# and tried to create a text-based interface which is motivated by RH's  #
# setup utility.                                                         #
#                                                                        #
#                              DATA STRUCTURE                            #
#                                                                        #
# The data structure used to hold the questions is a hash of records,    #
# which are implemented as a secondary hash.  The important part to      #
# understand well is the parent-child relationships between records.     #
# Each question has one or two children.  If it is a Yes/No question and #
# there are questions that should only be asked if the answer was, say,  #
# Yes, then the yes_child will differ from the no_child.  Each question  #
# has a proper_parent item, though.  The easiest way to make sense of    #
# this is by example:                                                    #
#                                                                        #
#  Q1:  Do you wanna answer IPCHAINS questions?                          #
#            yes_child= Q1a   no_child= Q2                               #
#                                                                        #
#     Q1a: ipchains question...  yes_child=no_child=Q1b   parent= Q1     #
#     Q1b: ipchains question 2   yes_child=no_child=Q2    parent= Q1a    #
#                                                                        #
#  Q2:  Wanna talk about sendmail?  proper_parent=Q1                     #
#                                                                        #
#                                                                        #
# OK, so the data structure goes like this...  Each question is indexed  #
# in the outer hash by a short name.  Remember, we're dealing with a     #
# hash of references to hashes, which behaves (thanks Larry) like a      #
# hash of hashes.  If you don't know what a reference is, don't worry    #
# as Perl kinda does something pretty intuitive here (thanks Larry).     #
# Within each question are the following fields:                         #
#                                                                        #
#                              RECORD FIELDS                             #
#                                                                        #
# question        --  question text (ex: wanna do this?)                 #
# short_exp       --  explanation text, at lower detail.                 #
# long_exp        --  explanation text, at high detail. (optional)       #
# default_ans     --  a default answer for the question (optional)       #
# toggle_yn       --  1 if a Yes/No question, 0 if fill-in-the-blank     #
# yes_epilogue    --  epilogue to display after question is asked (opt)  #
# no_epilogue     --  same as above, but displayed if this is y/n        #
#                     question and selected answer was no (optional)     #
# yes_child       --  index of next question, if this is not a Y/N ques. #
#                     or if this is a Y/N ques. and answer=Y             #
# no_child        --  index of next question, if Y/N ques. and answer=N  #
# proper_parent   --  index of "proper" previous question.  This is      #
#                     explained above, but, to recap: this is used so    #
#                     1+-> 1a -> 1b -> 1c -> 2, while 2's parent is 1    # 
#                      +-> 2                                             #
#                                                                        #
#                     where threading is due to 1 being a Y/N question   #
#                                                                        #
# confirm_text    --  this is text added to the end of an answer to      #
#                     maintain compatibility with the 1.0-1.2 ipchains   #
#                     script, which requires that most (but not all) of  #
#                     the user's answers be confirmed with a Y\n         #
#                                                                        #
# answer          --  answer chosen -- this is initially populated by    #
#                     the default_answer field's contents                #
# module          --  associated Bastille modules (automatically set     #
#                         by Load_Questions)                              #
#                                                                        #
#                                                                        #
#                              BASIC OPERATION                           #
#                                                                        #
# The TUI is basically composed of three big subroutines:                #
#                                                                        #
#   Load_Questions -- reads in the question data structure from a flat   #
#                     text file (Questions.txt) and does some sanity     #
#                     checking.                                          #
#                                                                        #
#   Ask            -- goes through all the business of asking a question #
#                     including: calling subs to draw the screen,        #
#                     scroll the explanation, and take input.  It has    #
#                     logic to decide what question should be asked next #
#                     based on answers and returns the index of such     #
#                                                                        #
#   Output_Config_Files -- outputs the configuration files for the       #
#                          Bastille script:                              #
#                                                                        #
#     tui-generated-formatted-config -- formatted version, for later use #
#     tui-generated-raw-config       -- raw version, for current mods    #
#                                                                        #
#                                                                        #
#  For more detail, read the Source, Luke.                               #
#                                            - jjb                       #
#                                                                        #
##########################################################################

use Curses;
use Curses::Widgets;

# If this is being run in Test mode (first argument is a T), we only
# test the Questions database
if ($ARGV[0] =~ /^\s*T\s*$/) {
    $TEST_ONLY=1;
}

if ($ARGV[0] =~ /^\s*TnQ\s*$/) {
    $TEST_ONLY=0;
    $TEST_NO_QUIT=1;
}


# Later on, we have to require that we're run as root.
#if ($EUID != 0) {
#    print "You need to be root to run this program.\n";
#    die "Program not run as root...";
#}

# Number_Modules is the number of modules loaded in by Load_Questions
$Number_Modules=0;

#
# Highlighted button is the button currently chosen in the button bar
#     We preserve this from question to question...
#
$highlighted_button=1;

# These variables contain the state of and text for the detail button
$detail_toggle="high";
$detail_button{"low"}="< Explain More >";
$detail_button{"high"}="< Explain Less >";

# Load in the questions database
$Number_Modules=&Load_Questions;
#&Load_Questions_Stub;

# Set up the Curses environment
$window = new Curses;
start_color;

# Increment determines spacing between buttons
$increment=$COLS/4;

# Get to the real work of asking questions...
&Draw_Title;
# Draw instructions
# $window->addstr(24,25,"TAB to switch between windows");
# $window->refresh;

$next_question_index="Title_Screen";
while ($next_question_index ne "RUN_SCRIPT") {
    $next_question_index=&Ask($next_question_index);    
}

# Display Credits

open CREDITS,"CREDITS";
while (my $creditline=<CREDITS>) {
    $credits .= $creditline;
}
close CREDITS;
$window->erase;

txt_field( 'window'       => $window,
	   'ypos'         => 0,
	   'xpos'         => 0,
	   'lines'        => $LINES - 2, 
	   'cols'         => $COLS - 2,
	   'content'      => $credits,
	   'border'       => 'blue',
	   'draw_only'    => 0,
	   'title'        => "Bastille Linux Credits",
	   'pos'          => 1,
	   );

endwin;

# Output answers to the script and display

&Output_Config_Files;

# Run Bastille

#&Run_Bastille_with_Config;

exit;




sub Load_Questions {
# sub Load_Questions creates a data structure called %%Questions

###############################################################################
# &Load_Questions does:
#
# 1) Create a question record for the Title screen: no question, no default
#    answer, toggle_yn=0, just a Short Explanation=Title Screen
# 2) Load in each question, one by one, by grabbing the expected records one 
#    by one from a file.  Records, within the file, are described below.  
# 3) Append a "Done now--gonna run the script" screen as Y/N question...  Needs
#    to have Yes-Child to be Bastille_Finish  QuestionName:  End_Screen
#
# Record format within Questions file:
# - A record is terminated by a blank line which is not part of a quoted
#   string.
# - A new record is began by the LABEL: tag, along with the index for the
#   question.
# - Otherwise, the entries within the record can be in any order at all.  
# - Multi-line fields must be quoted in double-quotes..
# - Double quotes are allowed inside a string, but must be escaped, like \".
#
# WEIRD CASES:
#
#  1 If a record has no short_exp explanation, none is shown.  This is bad?
#  2 If a record has no question, no question is asked, but the explanation
#    is still displayed.  If this is the case, the default_answer is still
#    entered into the configuration, if it exists.
#  3 If a question has no answer, it doesn't create any blank lines or such
#    in the output file, as it will be skipped in &Output_Config_Files.  For
#    this reason, &Prompt_Input, which is only called when the record contains
#    a user-answerable question, pads a space to the end of any 0-length input.
#    Not to worry: Output_Config_Files replaces said space with 0-length input
#    NOTE: we couldn't just only print the answer field when a real question
#          existed -- this would improperly handle case 2.
#    
###############################################################################


    # Line loaded in from file and it's non-tag data
    my ($line,$data);
    
    # Line number (within the disk file)
    my $line_number=0;

    # Module to which the current question being loaded in belongs and
    # the order that Load_Questions loads it in
    my $current_module;
    my $current_module_number=0;

    # Question (record) that we're in, listed by index(LABEL) 
    my $current_index;

    # The first and last questions -- used for Title_Screen and End_Screen 
    my $first_question="";
    my $previous_question="";

    # Field we're in
    my $current_field;

    
    # Set up Title Screen
    $Question{"Title_Screen"}{"question"}="";
    $Question{"Title_Screen"}{"short_exp"}=<<END_TITLESCREEN;
                            (Text User Interface)                  

				 v1.0.4.pre4
			          

       Please answer all the questions to build a more secure system.
       You can use the TAB key to switch among major screen functions,
       like each question's explanation area, input area and button area.
       Within each of the three major areas, use the arrow keys to scroll
       text or switch buttons.

       Please address bug reports and suggestions to jay\@bastille-linux.org

END_TITLESCREEN

    $Question{"Title_Screen"}{"proper_parent"}="Title_Screen";
    # Pay particularly good attention here! Title_Screen must have the index
    # for the first question...  Load this from the file later!
    $Question{"Title_Screen"}{"yes_child"}="FakeQ1";

    #
    # Load in the questions file, already!
    #

    # OK, so here's how this goes.  The config file looks has a series of
    # records, which have a series of  TAG: value    lines.  A value may span
    # multiple lines if it was begun with a  "  mark, but is usually expected
    # to be a string.  " marks can occur inside records, if escaped:  \"
    # Records end with a blank line and begin with a LABEL: tag.  We get module
    # names from FILE: somename.pm  lines, which should have whitespace around
    # them...

    open QUESTIONS,"./Questions.txt" or die "Can't open ./Questions.txt TUI data file...";
    print "Loading questions database...\n";
    while ($line=<QUESTIONS>) {

	# Increment line number
	$line_number++;

	# If we're currently in a question record...
	if ($current_index) {

	    # If we're currently in a __multi-line__ record (a quoted string),
	    if ($current_field) {

		# See if it's terminated in a quote (i.e. is end of a string?)
		my $end_of_string=0;

		if ($line =~ /^(.*)\"\s*\n*$/) {
		    # Make sure the terminating quote isn't an escaped quote
		    my $templine=$1;
		    unless ($templine =~ /\\$/) {
			$line=$templine;
			$end_of_string=1;
		    }
		}

		#
		### Text Handling
		#

		# Convert escaped quotes \" to real quotes "
		$line =~ s/\\\"/\"/g;

		# Strip out terminating \n's
		unless ($line =~ /^\s*\n+\s*$/) {
		    if ($line =~ /^(.*)\n$/) {
			$line = $1 . " ";
		    }
		}
		else {
		    $line .= "\n";
		}

		# Add the line to the end of the record and ...
		if ($Question{$current_index}{$current_field} =~ /[^\s\n]$/) {
		    $Question{$current_index}{$current_field} .= " ";
		}

		$Question{$current_index}{$current_field}.=$line;

		# Check if the record is over.
		if ($end_of_string) {
		    $current_field="";
	        }
            }
            else {
                # We're _not_ in a multi-line record
		
		# Did we hit a blank line? Blank lines, not embedded in
		#  " marks, delimit records
		if ($line =~/^\s*$/) {
		    $current_field="";
		    $current_index="";
		}
		else {
		    # Figure out what field to put this data in...
		    
		    if ($line =~ /^SHORT_EXP:(.*)$/) {
			$current_field="short_exp";
		    }
		    elsif ($line =~ /^LONG_EXP:(.*)$/) { 
			$current_field="long_exp";
		    }
		    elsif ($line =~ /^QUESTION:(.*)$/) {
			$current_field="question";
		    }
		    elsif ($line =~ /^DEFAULT_ANSWER:(.*)$/) {
			$current_field="answer";
		    }
		    # What I wouldn't give for a case/switch in Perl...
		    elsif ($line =~ /^YES_EXP:(.*)$/) {
			$current_field="yes_epilogue";
		    }
		    elsif ($line =~ /^NO_EXP:(.*)$/) {
			$current_field="no_epilogue";
		    }
		    elsif ($line =~ /^CONFIRM_TEXT:(.*)$/) {
			$current_field="confirm_text";
		    }
		    elsif ($line =~ /^YN_TOGGLE:(.*)$/) {
			$current_field="toggle_yn";
		    }
		    elsif ($line =~ /^YES_CHILD:(.*)$/) {
			$current_field="yes_child";
		    }
		    elsif ($line =~ /^NO_CHILD:(.*)$/) {
			$current_field="no_child";
		    }
		    elsif ($line =~ /^PROPER_PARENT:(.*)$/) {
			$current_field="proper_parent";
		    }
		    else {
			# UH OH!!! We've found a line (inside a record)
			# that isn't recognized
			print "ERRRRMMMM!!!!!\nThe following line (inside ";
			print "a record is not recognized!\n\n$line\n\n";
			print "Line number: $line_number\n";
			die "Line not recognized!";
		    }

		    # OK, we've know out what field to assign this data to...
		    $data =$1;

		    # Strip off any initiating white space
		    if ($data =~ /^\s+(.*)$/) {
			$data=$1;
		    }
		    
		    # Strip off any terminating white space...
		    if ($data =~ /^(.*)\s+$/) {
			$data=$1;
		    }
		    
		    # If the data isn't quoted, just finish up...
		    unless ( ($data =~ /^\"/ ) or ($data =~ /[^\\]\"$/) ){

			# Convert escaped quote marks
			$data =~ s/\\\"/\"/g;

			$Question{$current_index}{$current_field}=$data;

			$current_field="";
		    }
		    else {
			# Make sure this looks like a real quoted string
			if ($data !~ /^\s*\"/) {
			    print "ERROR: Weird quoted line\n\n$line\n\n";
			    print "Line number $line_number\n";
			    die "quoted line without quote at beginning of line";
			}

			# Strip off initiating quote mark
		        if ($data =~ /^\s*\"(.*)$/) {
			    $data=$1;
			}

			# If this thing has a terminating quote mark, it is a
			# single-line quoted record, probably used to preserve
			# leading or trailling whitespace.
			if ($data =~ /[^\\]\"$/) {
			    
			    if ($data =~ /^(.*)([^\\])\"$/) {
				$data=$1 . $2;
				
				# Convert escaped quote marks
				$data =~ s/\\\"/\"/g;

				$Question{$current_index}{$current_field}=$data;
				$current_field="";
			    }
			}
			# Otherwise, it is the beginning of a multi-line record
			else {

			    # Strip off end \n's
			    unless ($data =~ /^\s*\n+\s*$/) {
				if ($data =~ /^(.*)\n$/) {
				    $data=$1 . " ";
				}
			    }
			    
			    # Convert escaped quote marks
			    $data =~ s/\\\"/\"/g;

			    $Question{$current_index}{$current_field}=$data;
			}
		    }		 
		}
	    }
	}
	# OK, so we're not in a record (Question) at all...
	else {

	    # Are we starting a new one, on another blank line, or getting a
	    # module name?
	    if ($line =~ /^LABEL:\s*(.*)$/) {

		# We have found a new record...
		$current_index=$1;

		# Prune whitespace from the name
		if ($current_index=~/^(.*)\s+$/) {
		    $current_index =$1;
		}

		$Question{$current_index}{"module"}=$current_module;
		$current_field="";  # This is not a multi-line record

                # Record the name of the record so sanity checks can be done
		# later
		$recordnames[@recordnames]=$current_index;
		
		# If this is the first record (question), make the necessary
		# link from Title_Screen record
		unless ( $first_question ) {
		    $first_question=$current_index;
		}
		
		# Save the value of the current index so the End_Screen can
		# find the right parent...
		$previous_question=$current_index;

	    }
	    elsif ($line =~ /^FILE:\s*(.*)$/) {
		# Started a new module name...
		$current_module_number++;
		$current_module=$1 . " Module $current_module_number";
	    }
	    elsif ($line =~ /^\s*$/) {
		# A blank line
		# do nothing
	    }
	    else {
		# Something unexpected!
		print "Something weird discovered between question records:\n";
		print "$line\n\n";
		print "Line number $line\n";
		die "Problem with question file -- something strange found between question records!";
	    }
	}
    }
    close QUESTIONS;


    #
    # Set up Title_Screen and last question: "Can we run now?" screen 
    #

    $Question{"Title_Screen"}{"yes_child"}=$first_question;
    $Question{$first_question}{"proper_parent"}="Title_Screen";
    $Question{"End_Screen"}{"proper_parent"}=$previous_question;
    $Question{$previous_question}{"yes_child"}="End_Screen";
    $Question{$previous_question}{"no_child"}="End_Screen";

    $Question{"End_Screen"}{"short_exp"}="We will now implement the choices you have chosen here.\n\nAnswer NO if you want to go back and make changes!\n";
    $Question{"End_Screen"}{"question"}="Are you finished answering the questions, i.e. may we make the changes?";
    $Question{"End_Screen"}{"toggle_yn"}=1;
    $Question{"End_Screen"}{"yes_child"}="RUN_SCRIPT";
    $Question{"End_Screen"}{"no_child"}="End_Screen";
    $Question{"End_Screen"}{"no_epilogue"}= <<END_BACK_NEXT;
Please use Back/Next buttons to move among the questions you wish to change.


      Choose YES on this question later to implement your choices.
END_BACK_NEXT

    ##############################################
    #   Run sanity checks on questions database  #
    ##############################################

    my $key;
    foreach $key (@recordnames) {

	my $parent,$yes_child,$no_child;
	
	$parent=$Question{$key}{"proper_parent"};
	$yes_child=$Question{$key}{"yes_child"};
	if ($Question{$key}{"toggle_yn"}) {
	    $no_child=$Question{$key}{"no_child"};
	}

	if ($TEST_ONLY) {
	    print "\nChecking key $key\n";
	    print "Yes-child $yes_child\n";	    
	    print "No-child no_child\n";
	    print "Proper_parent $parent\n";
	
	    print "-"x10 . "Spewing record $key\n";
	    print $Question{$key}{"short_exp"};
	    print "\nlong_exp:\n";
	    print $Question{$key}{"long_exp"};
	    print "\nquestion:\n";
	    print $Question{$key}{"question"};
	    print "\nanswer\n";
	    print $Question{$key}{"default_answer"};
	    print "\nyes_child\n";
	    print $Question{$key}{"yes_child"};
	    print "\nno_child\n";
	    print $Question{$key}{"no_child"};
	    print "\nparent\n";
	    print $Question{$key}{"proper_parent"};
	    
	    print "\n\n";
	}

	die "Question database corrupt! $key: no parent!" unless $parent;
	die "Question database corrupt -- $key \'s parent doesn't exist" unless (defined($Question{$parent}));
	die "Question database corrupt -- $key \'s yes_child doesn't exist" unless (defined($Question{$yes_child}));

	die "Question database corrupt! $key: no yes-child" unless $yes_child;
	if ($Question{$key}{"toggle_yn"}) {
	    die "Question database corrupt! $key: no no_child on y/n question" unless ($no_child);
	    die "Question database corrupt -- $key \'s no_child doesn't exist" unless (defined ($Question{$no_child}));
	    die "Question database corrupt -- y/n question $key has no Question!" unless ( $Question{$key}{"question"} );

	}


    }

    if ($TEST_ONLY) {
	exit;
    }      

    # Return number of modules loaded in
    $current_module_number;

}




sub Load_Questions_Stub {
# sub Load_Questions creates a data structure called %%Questions
    
    # Set up Title Screen
    $Question{"Title_Screen"}{"question"}="";
    $Question{"Title_Screen"}{"short_exp"}=<<END_TITLESCREEN;
                                Bastille Linux                          

	       	      (Text User Interface ver 1.00pre1)                  

			       SPECIAL TEST MODE



			            (jjb)


			         ( PRESS TAB )
END_TITLESCREEN

    $Question{"Title_Screen"}{"yes_child"}="FakeQ1";
    $Question{"Title_Screen"}{"proper_parent"}="Title_Screen";

    # Create a fake question 1 for testing purposes
    $Question{"FakeQ1"}{"short_exp"}="This is the\nfake question\n";
    $Question{"FakeQ1"}{"long_exp"}="This\nis\nthe\nlong version\nof the\nfake question.";
    $Question{"FakeQ1"}{"question"}="Do you want to do this?";
    $Question{"FakeQ1"}{"yes_child"}="FakeQ2";
    $Question{"FakeQ1"}{"proper_parent"}="Title_Screen";
    $Question{"FakeQ1"}{"yes_epilogue"}="OK, we did it";
    $Question{"FakeQ1"}{"toggle_confirm"}="1";

    $Question{"FakeQ2"}{"short_exp"}= <<END_DETAIL_SETTING;
We will be setting the default level of detail of the explanations to Longer Explanations.  You can toggle this, as many times as you like, at each question by TAB-bing over to the DETAIL button and hitting Enter/Space.

Please select NEXT to go to the next question.
END_DETAIL_SETTING

    $Question{"FakeQ2"}{"answer"}="Y";
    $Question{"FakeQ2"}{"proper_parent"}="FakeQ1";
    $Question{"FakeQ2"}{"yes_child"}="End_Screen";

    # Set up last question: "Can we run now?" screen -- careful about Parent!!!
    $Question{"End_Screen"}{"proper_parent"}="FakeQ2";
    $Question{"End_Screen"}{"short_exp"}="We will now implement the choices you have chosen here.\n\nAnswer NO if you want to go back and make changes!\n";
    $Question{"End_Screen"}{"question"}="Are you finished answering the questions, i.e. may we make the changes?";
    $Question{"End_Screen"}{"toggle_yn"}=1;
    $Question{"End_Screen"}{"yes_child"}="RUN_SCRIPT";
    $Question{"End_Screen"}{"no_child"}="End_Screen";
    $Question{"End_Screen"}{"no_epilogue"}="Please use Back/Next buttons to move among the questions you wish to\nchange.\n\nChoose YES on this question later to implement your choices.\n";

    
    1;
}


sub Ask {
# sub Ask (index to Question{} record)

##############################################################################
#
# More comments coming here, from Design Doc...
#
# WEIRD CASES:
#
#  1 If a record has no short_exp explanation, none is shown.  This is bad?
#  2 If a record has no question, no question is asked, but the explanation
#    is still displayed.  If this is the case, the default_answer is still
#    entered into the configuration, if it exists.
#  3 If a question has no answer, it doesn't create any blank lines or such
#    in the output file, as it will be skipped in &Output_Config_Files.  For
#    this reason, &Prompt_Input, which is only called when the record contains
#    a user-answerable question, pads a space to the end of any 0-length input.
#    Not to worry: Output_Config_Files replaces said space with 0-length input
#    NOTE: we couldn't just only print the answer field when a real question
#          existed -- this would improperly handle case 2.
#    
###############################################################################

    #
    # Load question into local variables
    #
    my $index=$_[0];
    my $short_exp      =$Question{$index}{"short_exp"};
    my $long_exp       =$Question{$index}{"long_exp"};
    my $question       =$Question{$index}{"question"};
    my $toggle_confirm = $Question{$index}{"toggle_confirm"};
    my $toggle_yn      =$Question{$index}{"toggle_yn"};
    my $yes_epilogue   =$Question{$index}{"yes_epilogue"};
    my $no_epilogue    =$Question{$index}{"no_epilogue"};
    my $yes_child      =$Question{$index}{"yes_child"};
    my $no_child       =$Question{$index}{"no_child"};
    my $proper_parent  =$Question{$index}{"proper_parent"};
    my $answer         =$Question{$index}{"answer"};
    my $module         =$Question{$index}{"module"};
    
    #
    # Set up other local variables
    #

    # Index to next question -- returned by this routine -- loop until we
    # have an index
    my $return_index="";

    # Highlighted button stores the button currently hightlighted -- set this
    # to 1 whenever we enter a question
    # my $highlighted_button=1;


    # Widgets 1.1 feature
    # Current position in the Scrollable explanation box
    #my $current_scroll_position=0;

    while (not $return_index) {

	# Button pressed at end
	my $button_pressed="";
	
	# Input window exited via enter key instead of tab key
	my $exited_with_enter=0;

	#
	# Explanation to display in routines -- we use short explanation if 
	# long is empty, since long explanation is optional
	
	my $explanation;
	if (($detail_toggle eq "high") and ($long_exp)) {
	    $explanation="$long_exp\n\n$question";
	}
	elsif ($short_exp) {
	    $explanation="$short_exp\n\n$question";
	}
	else {
	    # Special case: when there is no explanation, this is a question
	    # used either to show an epilogue or to (more likely) get some
	    # default answer into the data stream

	    $explanation="";
	}

	#
	# Draw the screen before accepting input
	#
	# This basically does the following, making it easier to continue
	# redrawing the screen on every window switch...
	# &Draw_Explanation($explanation);
	# if ($question) { 
	#     &Draw_Prompt($answer,$toggle_yn);
	# }
	# &Draw_Buttons;

	&Redraw_Screen($module,$explanation,$question,$answer,$toggle_yn);

	#
	# Redraw windows with user input allowed and accept input
	#

	# Change from 1.0.4.pre1 -> 1.0.4.pre2  -- switched window order
	#if ($explanation) {
	#    &Scroll_Explanation($module,$explanation);
	#    &Redraw_Screen($module,$explanation,$question,$answer,$toggle_yn);
	#}
	if ($question) {
	    ($answer,$exited_with_enter)=&Prompt_Input($answer,$toggle_yn);
	    &Redraw_Screen($module,$explanation,$question,$answer,$toggle_yn);
	}
	else {
	    if ($answer) {
		$button_pressed="Next";
	    }
	}
	
	# Skip the buttons if we exited the input with the Enter key
	if ($exited_with_enter) {
	    $button_pressed="Next";
	}
	elsif ($question or $explanation)  {
	    $button_pressed=&Allow_Buttons;
	    &Redraw_Screen($module,$explanation,$question,$answer,$toggle_yn);
	}

	# Was the detail level of the explanations toggled?
	if ($button_pressed eq "Detail") {
	    # Switch the detail level
	    if ($detail_toggle eq "high") {
		$detail_toggle="low";
	    }
	    else {
		$detail_toggle="high";
	    }

	    # Reload the explanation
	    if (($detail_toggle eq "high") and ($long_exp)) {
		$explanation="$long_exp\n\n$question";
	    }
	    elsif ($short_exp) {
		$explanation="$short_exp\n\n$question";
	    }

	}
	
	# Did the user hit the Back Button?
	if ($button_pressed eq "Back") {
	    $return_index=$proper_parent;
	}

	# Did the user hit the next button on an answered question?
	# Ahhh, we allow blank responses to questions -- make this less
	# icky later by using radio buttons/lists for yes/no...
	#
#	if (($button_pressed eq "Next") and ($answer)) {
	if  ($button_pressed eq "Next") {

	    # If not a Y/N question, use the Yes Child and Epilogue
	    unless ($toggle_yn) {		
		$return_index=$yes_child;
		
		if ($yes_epilogue) {
		    $window->erase;
		    &Draw_Title;
		    &Display_Epilogue($yes_epilogue);
		}
	    }
	    # Otherwise, use the epilogue appropriate to the answer
	    else {
		if ($answer=~/^\s*Y/i) {
		    $return_index=$yes_child;

		    $window->erase;
		    &Draw_Title;
		    &Display_Epilogue($yes_epilogue);
		}
		elsif ($answer =~/^\s*N/i) {
		    $window->erase;
		    &Draw_Title;
		    $return_index=$no_child;
		    &Display_Epilogue($no_epilogue);
		}
	    }
	}

	# Now, if the user hit Prev/Next, store the answer so it's there
	# when we come back to it...
	if ($return_index) {
	    if ($answer) {
		$Question{$index}{"answer"}=$answer;
	    }
	}
	else {
	    &Scroll_Explanation($module,$explanation);
	    &Redraw_Screen($module,$explanation,$question,$answer,$toggle_yn);
	}

	# Deal with text-screen questions: questions w/ no question and
        # explanation which still have an epilogue

	if ((not $explanation) and (not $question)) {
	    $return_index=$yes_child;
	    &Display_Epilogue($yes_epilogue);
	}

    }
		
    $return_index;
}

sub Redraw_Screen {
# sub Redraw_Screen($module,$explanation,$question,$answer,$toggle_yn)
    
    my ($module,$explanation,$question,$answer,$toggle_yn)=@_;

    # Erase the screen.
    $window->erase;

    &Draw_Title;
    &Draw_Key_Instructions;

    if ($explanation) {
	&Draw_Explanation($module,$explanation);
    }
    if ($question) { 
	&Draw_Prompt($answer,$toggle_yn);
    }
    if ($question or $explanation) {
	&Draw_Buttons;
    }
}

# Window drawing functions...

sub Draw_Title {
# Draw the title bar   

    txt_field( 'window' => $window, 
	       'ypos'   => 0, 'xpos'   => 0,
	       'lines'  => 1,
	       'cols'   => $COLS-2, 
	       'content'=> "                                Bastille Linux",
	       'border' => 'blue', 
	       'draw_only' => 1,
	       );
}

sub Draw_Key_Instructions {

    select_colour($window,"yellow");
    $window->addstr(24,0,"TAB to switch Windows                               Arrow Keys to switch Buttons");
    $window->attrset(0);

    1;
}
    
    

sub Draw_Explanation {
# Draw the explanation

    my ($module,$explanation)=@_;
    my $title;

    if ($module) {
	$title=$module . " of $Number_Modules";
    }

    txt_field( 'window'       => $window,
	       'ypos'         => 3,
	       'xpos'         => 0, # $COLS - 5,
	       'lines'        => $LINES - 11, 
	       'cols'         => $COLS - 2,
	       'content'      => $explanation,
	       'border'       => 'blue',
	       'draw_only'    => 1,
	       'title'        => $title,
	       'pos'          => 0,
# Widgets 1.1 feature
#	       'pos'          => $current_scroll_position,
	       # 'function'     => \&clock
	       );
}

sub Draw_Prompt {
# Draw the Answer area

    my ($current_answer,$toggle_yn)=@_;    
    my %yes_no_list =("0","Yes","1","No");
    my $default_entry=0;

    # Is this a yes/no question or not? If yes, use a button list...
    if ($toggle_yn) {
	if ($current_answer=~/^\s*Y/i) {
	    $default_entry=0;
	}
	elsif ($current_answer=~/^\s*N/i) {
	    $default_entry=1;
	}
	    
	my ($key, $answer)=list_box( 'window'    => $window,
				     'ypos'     => $LINES-6,
				     'xpos'     => 36,
				     'lines'     => 2,
				     'cols'      => 5,
				     'border'    => 'blue',
				     'selected'  => $default_entry,
				     'list'      => \%yes_no_list,
				     'draw_only' => 0
				     );

    }
    # If it's not a yes/no question, create an answer blank...
    else {

	txt_field( 'window'        => $window,
		   'ypos'          => $LINES-6,
		   'xpos'          => 1,
		   'lines'         => 2,
		   'cols'          => $COLS-4,
		   'content'       => "Answer: $current_answer",
		   'border'        => 'blue',
		   'draw_only'     => 1,
		   'pos'           => 8,
		   );
    }
}

sub Draw_Buttons {

# Use of the widget removed for now, since draw_only doesn't work here
#($input,$selected) = buttons(    'window'       => $window,
#                                 'buttons'       => ["Back","Next",
#					$detail_button{$detail_toggle}],
#				 'ypos'          => $LINES-2,
#				 'xpos'          => ($increment)-2,
#				 'spacing'       => $increment/4,
#				 'active_button' => 1,
#				 'draw-only'     => 1,
#				 );
# Emulate!
    select_colour($window,"blue");
    $window->addstr($LINES-2,($increment)-4,"< Back >");
    $window->addstr($LINES-2,($increment*2)-11,"< Next >");
    $window->addstr($LINES-2,($increment*3)-18,$detail_button{$detail_toggle});
    $window->attrset(0);
    $window->refresh;
}

sub Scroll_Explanation {
# Allow user to scroll the explanation

    my ($module,$explanation)=@_;
    my $title;

    if ($module) {
	$title=$module . " of $Number_Modules";
    }

    noecho;
    # Widgets 1.1 feature
    #($key, $rtrnd_note,$current_scroll_position) = 
    ($key,$rtrnd_note) =  txt_field( 'window'       => $window,
				     'ypos'         => 3,
				     'xpos'         => 0, # $COLS - 5,
				     'lines'        => $LINES - 11, 
				     'cols'         => $COLS - 2,
				     'content'      => $explanation,
				     'border'       => 'red',
				     'draw_only'    => 0,
                                     # Widgets 1.1 only -- no edit in window
				     #'edit'         => 0,
				     'title'        => $title,
				     'pos'          => $current_scroll_position,
				     # 'function'     => \&clock
				     );
}

sub Display_Epilogue {
    ###
#######  See about using text widget, but blanking the buttons and answer for
#######  this...
    ###

    my $text=$_[0];
    
    if ($text) {
	$text = " "x27 . "<Press TAB to go on>\n\n" . $text;

	($key, $rtrnd_note) = txt_field( 'window'       => $window,
					 'ypos'         => 3,
					 'xpos'         => 0,
					 'lines'        => $LINES-5,
					 'cols'         => $COLS-2,
					 'content'      => $text,
					 'border'       => 'red',
					 'pos'          => 1,
					 );
#    	msg_box('message' => $text);
    }
}

sub Prompt_Input {
# sub Prompt_Input($current_answer, $toggle_yn)
#
# $current_answer -- the answer to place in the window for user to edit
# $toggle_yn      -- 1 if this is a y/n question, 0 otherwise

###############################################################################
# WEIRD CASES:
#
#  1 If a record has no short_exp explanation, none is shown.  This is bad?
#  2 If a record has no question, no question is asked, but the explanation
#    is still displayed.  If this is the case, the default_answer is still
#    entered into the configuration, if it exists.
#  3 If a question has no answer, it doesn't create any blank lines or such
#    in the output file, as it will be skipped in &Output_Config_Files.  For
#    this reason, &Prompt_Input, which is only called when the record contains
#    a user-answerable question, pads a space to the end of any 0-length input.
#    Not to worry: Output_Config_Files replaces said space with 0-length input
#    NOTE: we couldn't just only print the answer field when a real question
#          existed -- this would improperly handle case 2.
#    
###############################################################################

# Allow user to input an answer

    my ($current_answer,$toggle_yn)=@_;

    # Default item to have highlighted
    my $default_entry=0;

    # List of buttons for the y/n listbox widget
    my %yes_no_list =(0,"Yes",1,"No");

    # Input variables for the widgets
    my ($key,$answer);

    # Used to keep track of how widget was exited
    my $exited_with_enter=0;

    # Is this a yes/no question or not? If yes, use a button list...
    if ($toggle_yn) {
	if ($current_answer=~/^\s*Y/i) {
	    $default_entry="0";
	}
	elsif ($current_answer=~/^\s*N/i) {
	    $default_entry="1";
	}
	    
	($key, $answer)=list_box(    'window'    => $window,
				     'ypos'      => $LINES-6,
				     'xpos'      => 36,
				     'lines'     => 2,
				     'cols'      => 5,
				     'border'    => 'red',
				     'selected'  => $default_entry,
				     'list'      => \%yes_no_list,
				     );
	if ($answer==0) {
	    $answer="Y";
	}
	else {
	    $answer="N";
	}

	if ($key eq "\n") {
	    $exited_with_enter=1;
	}
    }
    # If it's not a yes/no question, create an answer blank...
    else {
	($key, $answer) = txt_field( 'window'        => $window,
					'ypos'          => $LINES-6,
					'xpos'          => 1,
					'lines'         => 2,
					'cols'          => $COLS-4,
					'content'       => "Answer: $current_answer",
					'border'        => 'red',
					'draw_only'     => 0,
					'pos'           => 8,
					);

	# If exited with enter, set flag
	if ($key eq "\n") {
	    $exited_with_enter=1;
	}

	# Prune off the prompt which is retained by the txt_field widget
	if ($answer =~ /^Answer:\s*(.*)/) {
	    $answer=$1;
	}

	# If answer is only a comment, blank it.
	if ($answer =~ /^\s*\#/ ) {
	    $answer=" ";
	}

	# If the $answer field is zero-length, add a space...  Read about
	# this above...

	if ($answer =~ /^$/) {
	    $answer=" ";
	}

	# If the entered answer is multi-lined, this won't do...  We prune
	# out all but the first non-blank line (plus any comment lines)
#	if ($answer =~ /\n/) {
#	    my $line,$pruned_answer;
#	    foreach $line (split($answer,)) {
#		if ($line =~ 
#
#	    }
#	}
    }

    return ($answer,$exited_with_enter);
}

sub Allow_Buttons {

    my ($input,$selected);

    while ($input !~ /[\t \n]/) {
	($input,$selected) = buttons(    'window'       => $window,
				     'buttons'       => ["< Back >","< Next >",
					       $detail_button{$detail_toggle}],
				     'ypos'          => $LINES-2,
				     'xpos'          => ($increment)-4,
				     'spacing'       => $increment/4,
				     'active_button' => $highlighted_button,
				     'draw-only'     => 0,
				 );
    }

    $highlighted_button=$selected;

    if ($selected == 2) {
	$selected="Detail";
    }
    elsif ($selected == 0) {
	$selected="Back";
    }
    else {
	$selected="Next";
    }

    # But wait! If we exited button widget via Tab, return none of these b/c
    # the user didn't SELECT a button!
    if ($input eq "\t") {
	$selected="Repeat_Scroll";
    }

    $selected;
}


sub Output_Config_Files {

##############################################################################
# %Output_Answers will:
#
# 1) write out a file, which can be read by AppConfig, containing the full
#    contents of the questions data structure
#
# 2) write out a "raw" answer-file which can be fed into the current Bastille
#    script via AutomateBastille or the front-end
#
# More information coming here, from design doc...
#
# WEIRD CASES:
#
#  1 If a record has no short_exp explanation, none is shown.  This is bad?
#  2 If a record has no question, no question is asked, but the explanation
#    is still displayed.  If this is the case, the default_answer is still
#    entered into the configuration, if it exists.
#  3 If a question has no answer, it doesn't create any blank lines or such
#    in the output file, as it will be skipped in &Output_Config_Files.  For
#    this reason, &Prompt_Input, which is only called when the record contains
#    a user-answerable question, pads a space to the end of any 0-length input.
#    Not to worry: Output_Config_Files replaces said space with 0-length input
#    NOTE: we couldn't just only print the answer field when a real question
#          existed -- this would improperly handle case 2.
#    
##############################################################################

    open RAW_CONFIG,"> tui-generated-raw-config";
    open FORMATTED_CONFIG,"> tui-generated-formatted-config";

    my $index="Title_Screen";
    while ($index ne "End_Screen") {
	if ($Question{$index}{"answer"}) {

	    # If the answer is just a space (the way the &Prompt_Input sub
	    # designates a blank line, strip it.
	    
	    if ($Question{$index}{"answer"} =~ /^\s$/) {
		$Question{$index}{"answer"} = "";
	    }

	    # Print the raw input for the Bastille 1.0 script series 
	    print RAW_CONFIG "### Question index $index\n";
	    if ($Question{$index}{"question"}) {
		print RAW_CONFIG "### Question text  ";
		print RAW_CONFIG $Question{$index}{"question"} . "\n";
	    }

	    # The "confirm_text" data element fixes a problem caused by the
	    # different way that IPChains handles input.  In the IPCHAINS 
	    # module, there are two critical items:
	    # 
            #  1) if the user enters an empty string (by hitting enter),
	    #     IPCHAINS replaces this with the default answer.  A true
	    #     "empty" result reqires that the user enter 
	    #     something (ie, a space)
	    #  2) Most (not _all_) of the questions require a Y\n confirmation
	    #
	    # By including confirmation text, we can kill the first difference
	    # by having the confirm text include a " ".  Should the question
	    # also need a confirming Y, this can be included by making the
	    # confirmation text " \nY".
	    # Note: to make this work, we have to fix \n's...
	    
	    print RAW_CONFIG $Question{$index}{"answer"};
	    $Question{$index}{"confirm_text"} =~ s/\\n/\n/g;
	    print RAW_CONFIG $Question{$index}{"confirm_text"};
	    print RAW_CONFIG "\n";

	    # Print formatted file too...
	    my $module = $Question{$index}{"module"};
	    if ($module =~ /^([^.]+).pm/) {
		$module =$1;
	    }
	    print FORMATTED_CONFIG  "$module.";
	    print FORMATTED_CONFIG $index . "=\"";
	    print FORMATTED_CONFIG $Question{$index}{"answer"} . "\"\n";
	    
	}
	if ($Question{$index}{"toggle_yn"} == 0) {
	    $index=$Question{$index}{"yes_child"};
	}
	else {
	    if ($Question{$index}{"answer"} =~ /^\s*Y/i) {
		$index=$Question{$index}{"yes_child"};
	    }
	    elsif ($Question{$index}{"answer"} =~ /^\s*N/i) {
		$index=$Question{$index}{"no_child"};
	    }
	    else {
		print RAW_CONFIG "Major error on question $index!\n";
		print "Major error on y/n question $index -- answer not y/n !\n";
	    }
	}
    }
    close FORMATTED_CONFIG;
    close RAW_CONFIG;
}



sub Run_Bastille_with_Config {
    
    system "./BackEnd.pl < tui-generated-raw-config > bastille-raw-output-log";
}



